# -*- coding: utf-8 -*-
"""
Created on Mon Jan 11 14:46:32 2021

@author: Guenole CHEROT
"""
import numpy as np
from copy import deepcopy

class Communication():
    """
    Communication between SO and Agents. Remembers :
        - Trades between agents
        - Network cost
        
    Input :
        - n_agent (int) : number of agent in the network
        - grid_cost (float) : grid cost imposed by the SO
    
    Attribute :
        - P_buffer (array) : list of the power that each agent wants to exanche at the end of their minimization
        - P (array) : Same as P_buffer but when all agents have converged
        """
    
    def __init__(self, n_agent, grid_cost):
        self.n_agent = n_agent
        self.P = np.zeros((n_agent, n_agent))
        self.P_buffer = np.zeros((n_agent, n_agent))
        self.grid_cost = grid_cost
        self.t = 0
        
    def save_network_fees(self, grid_cost):
        """Save the network cost given by the SO. \n
        Input :
            - grid_cost (float) : network cost imposed by SO
            """
        self.grid_cost = grid_cost
    
    def update_trade(self, trade, ID=None):
        """Update the trade proposed by the agent ID. \n
        Input : 
            - trade (array) : array size (n_agent) representing the trade proposed by the agent ID.
            - ID (int) : ID of the agent proposing the trade"""
        if ID == None:
            if trade.shape == (self.n_agent, self.n_agent):
                self.P_buffer = trade
            else :
                raise ValueError("trade as not the correct shape")
        else :
            self.P_buffer[ID] = trade
        
        
    def get_trade(self, ID=None):
        """Get the trade proposed by all agent to agent number ID. \n
        Input :
            - ID (int) : ID of the agent proposing the trade. If None return the trade matrix (P)
        Output :
            - trade (array) : array size (n_agent) representing the trade proposed by all agents to agent number ID."""
        if ID == None:
            return self.P
        else :
            return self.P[:,ID]
    
    def broadcast(self):
        """Update all value of power trade (used in a synchronous algorithm)"""
        # Update the power trade
        self.P = deepcopy(self.P_buffer)
        # Increment time
        self.t += 1


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 11 11:10:02 2021

@author: Guenole CHEROT
"""

import numpy as np
import warnings
from tqdm import tqdm
from joblib import Parallel, delayed
warnings.filterwarnings("ignore")

import pandapower as pp

from rt_congestion_control.memory import Memory
from rt_congestion_control.agent import Agent

class Admm():
    """Take all actors and perform the ADMM. \n
    Input :
        - Net : instance of the Network class panda_power
        - SO : instance of System opérator class
        - Communication : instance of the Communication class
        - rho (float) : Penalisation parameter of the ADMM
        - compute_opf (int, 1) : 0 does not compute OPF, 1 compute OPF with line limits, 2 compute OPF with and without line limits
        """
        
    ########################################################################################################
    ########################################### INIT and  SETERS ###########################################
    ########################################################################################################
    
    def __init__(self, net, SO, communication, rho=1000, compute_opf=1):
        self.net = net
        self.SO = SO
        self.communication = communication
        self.rho = rho
        self.compute_opf = compute_opf
        
        # Other usefull variables
        self.n_load = len(self.net.load)
        self.n_agent = self.n_load +1 #Add the ext grid
        self.n_step = len(self.net.data[self.net.data["dataid"]==1])
        
        # Create the agents
        self.create_agents(net.data)
        self.memory = self.init_memory(self.agents, SO)
        
        
        
    def create_agents(self, time_series):
        """Create all the agents (prosumer).
        Input :
            - time_series (DataFrame) : Contains the power asked by each agent trough time."""
        self.agents = []
        for k in self.net.flex_data.index:
            ID = self.net.flex_data.dataid[k]
            self.agents.append(Agent(name=None,
                                     ID=ID,
                                     element=ID,
                                     et=self.net.flex_data.et[k],
                                     bus=self.net.flex_data.bus[k],
                                     n_agent=self.n_agent,
                                     zone=self.net.flex_data.zone[k],
                                     flexibility=self.net.flex_data.flexibility[k],
                                     grid_cost_type = "uniq",
                                     rho=self.rho,
                                     partners="all",
                                     time_series=self.net.data[self.net.data["dataid"]==ID].sort_values(by='localminute')))
            
    def init_memory(self, agents, SO, n_step_warm=None):
        """Initialise a memory to store all data from the simulation.
        Can be used to store data from warm start.
        Input :
            - agets (list) : list of Agent
            - SO : instance of System opérator class
            - n_step_warm (int, None) : Number of iteration of the warm start, if the memory is used for warm start."""
        if n_step_warm == None :
            n_step = self.n_step
        else :
            n_step = n_step_warm
            
        memory = Memory(n_agent = self.n_load +1, #ext_grid is couted as a agent
                             n_step = n_step,
                             n_line = len(self.net.line),
                             n_bus = len(self.net.bus),
                             grid_cost_type=self.SO.grid_cost_type,
                             line_limits = self.SO.line_limits,
                             compute_opf=self.compute_opf,
                             data=self.net.data,
                             flex_data = self.net.flex_data)
        return memory
    
    ########################################################################################################
    ######################################## Network manipulations #########################################
    ########################################################################################################
        
    def update_network(self, trade, min_p_mw, max_p_mw, polycost):
        """Update the network with power aplied by agents.
        Input :
            - trade (array) : array of size (n_agent, n_agent) contaning all the trade computed by the agents
            - min_p_mw (array) : array of size (n_agent) representing the lower bound for each agent
            - max_p_mw (array) : array of size (n_agent) representing the upper bound for each agent
            - polycost (array) : array of size (n_agent, 3) representing the polynomial cost for each agent"""
        self.update_net_power(np.sum(trade, axis=1))
        self.update_bounds(min_p_mw, max_p_mw)
        self.update_polycost(polycost)
        
    def update_net_power(self, power):
        """Update the power produced by each agent in the panda_power network. Should be done after ADMM and before load flow. \n
        Input :
            - power (array) : vector size (n_agent) representing the power consumed by each agent."""
        # Make shure the total power balance is null (the exess is provided by the network)
        power[0] = -np.sum(power[1:]) # This line assumes that the first agent is the external grid
        
        # Change the sign because the power is alway positive in panda power
        for k in self.net.flex_data.loc[self.net.flex_data["et"]=="load"].dataid:
            self.net.load["p_mw"][k] = -power[k] # For loads : positive power means consumed
            

    def update_bounds(self, min_p_mw, max_p_mw):
        """Update the bounds of each agents. \n
        Input :
            - min_p_mw (array) : array of size (n_agent) representing the lower bound of each agent
            - max_p_mw (array) : array of size (n_agent) representing the upper bound of each agent"""
        for k in self.net.flex_data.index:
            ID = self.net.flex_data.dataid[k]
            if self.net.flex_data.et[k] == "load":
                self.net.load["min_p_mw"][ID] = min_p_mw[ID]
                self.net.load["max_p_mw"][ID] = max_p_mw[ID]
            elif self.net.flex_data.et[k] == "ext_grid":
                self.net.ext_grid["min_p_mw"][ID] = min_p_mw[ID]
                self.net.ext_grid["max_p_mw"][ID] = max_p_mw[ID]
    
    def update_polycost(self, polycost):
        """Update the polycost of each agents. \n
        Input :
            polycost (array) : array of size (n_agent, 3) representing the polynomial cost for each agent"""
        for k in range(self.n_agent):
            self.net.poly_cost.at[k,"cp1_eur_per_mw"] = polycost[k,1]
            self.net.poly_cost.at[k,"cp2_eur_per_mw2"] = polycost[k,2]
        
    def runpp(self, memory, t, grid_cost_array):
        """Run power flow and optimal flow in the network and return the results.
        Input :
            - memory (Memory class) : memory of the simulation.
            - t (int) : time of the simulation
            - grid_cost_array (array) : array of the grid cost paid by each agent (can by positive or negarive)
        return :
            - loading_percent : loading percent of each line after the powerflow
            - P_exchanged : power exchanged in the network (real power that transited)
            """
        try :
            pp.runpp(self.net, enforce_q_lims=True, algorithm="nr")
        except:
            print("PF did not converge")
        # save the network state
        loading_percent = self.net.res_line.sort_index()['loading_percent']
        P_exchanged = np.concatenate((self.net.res_ext_grid["p_mw"].to_numpy(), -self.net.res_load["p_mw"].to_numpy()))
        # Q_exchanged = np.concatenate((self.net.res_ext_grid["q_mvar"].to_numpy(), self.net.res_load["q_mvar"].to_numpy()))
        memory.line_loading_percent[t] = loading_percent
        memory.P_exchanged[t] = P_exchanged
        # memory.Q_exchanged[t] = Q_exchanged
        
        if self.compute_opf >= 1 :
            constraints  = [[self.net.line["max_loading_percent"][0], np.zeros(self.n_agent)]]
            if self.compute_opf == 2 :
                constraints.append([200, grid_cost_array])
            loading_percent_opf, P_exchanged_opf = zip(*Parallel(n_jobs=self.compute_opf)(delayed(self.run_opf)(c) for c in constraints))
            for k in range(self.compute_opf):
                memory.opf_results(loading_percent_opf[k], P_exchanged_opf[k], k, t)
        
        return loading_percent, self.net.res_ext_grid["p_mw"].to_numpy()[0]
    
    def run_opf(self, constraints):
        """Compute the OPF with the specified constraints.
        Input :
            - constraints (array) : Array of shape 2, the first argument is the line limits in %, the second the cost to be added in polycost."""
        max_loading = constraints[0]
        grid_cost = constraints[1]
        # Modify network line limits
        save_max_loading = self.net.line["max_loading_percent"] = max_loading
        self.net.line["max_loading_percent"] = max_loading
        # Modify polycost
        self.net.poly_cost.cp1_eur_per_mw += grid_cost
        try :
            pp.runopp(self.net, PDIPM_COMPTOL = 1e-12)
        except :
            print("OPF did not converge")
        # Set network how it was
        self.net.line["max_loading_percent"] = save_max_loading
        self.net.poly_cost.cp1_eur_per_mw -= grid_cost
        # Get results
        loading_percent_opf = self.net.res_line.sort_index()['loading_percent']
        P_exchanged_opf = np.concatenate((self.net.res_ext_grid["p_mw"].to_numpy(), self.net.res_gen["p_mw"].to_numpy(), -self.net.res_load["p_mw"].to_numpy()))
        # Q_exchanged_opf = np.concatenate((self.net.res_ext_grid["q_mvar"].to_numpy(), self.net.res_gen["q_mvar"].to_numpy(), self.net.res_load["q_mvar"].to_numpy()))
        return loading_percent_opf, P_exchanged_opf
    
    ########################################################################################################
    ############################################ Core functions ############################################
    ########################################################################################################
     
    def run(self, warm_start, max_it=500, SO_strategy = True, plot =False):
        """Run the simulation. \n
        Input :
            - warm_start (bool) : perform a warm start. If True the simulation start with the ADMM converged.
            - max_it (int, 500) : maximum number of itération
            - SO_strategy (True) : If True, the price can fluctuate during the warm start depending on the SO' strategy. Set False to have a fixed price.
            - plot (bool, False) : plot the convergence of the warm start
            """
        if warm_start==True:
            if SO_strategy == False :
                strategy = self.SO.strategy
                self.SO.strategy = "None" 
            self.memory_warm = self.init_memory(self.agents, self.SO, max_it)
            self.ADMM(self.memory_warm, warm_start = True, max_it=max_it)
            # Plot
            if plot ==True:
                self.memory_warm.plot_convergence(scale=True)
                self.memory_warm.plot_all_agent_power(t=max_it-1, warm=True)
                self.memory_warm.plot_agent_power_during_simu()
        
        if SO_strategy == False :
            self.SO.strategy = strategy
        self.ADMM(self.memory)
        
    def ADMM(self, memory, warm_start=False, max_it=None):
        """Apply the ADMM routine. \n
        Input :
            - memory : memory class that stores all data from the simulation
            - warm_start (bool) : Performs a warm start if True (max_it should be specified)
            - max_it (int) : only used for warm start, specifie the maximum number of iteration for the warm start.
            """
        # Init variable
        trade = np.zeros((self.n_agent, self.n_agent))
        min_p_mw = np.zeros(self.n_agent)
        max_p_mw = np.zeros(self.n_agent)
        polycost = np.zeros((self.n_agent,3))
        grid_cost_array = np.zeros(self.n_agent)
        
        # Init with warm start
        P = self.communication.get_trade()
        for k in range(len(self.agents)):
            self.agents[k].P1 = P[k,:]
        # Number of step
        if warm_start :
            n_step = max_it
        else :
            n_step = self.n_step
        # Progess bar
        pbar = tqdm(total=n_step)
        # Simulation
        for t in range(n_step):
            # Get trade and grid cost through communication
            P = self.communication.get_trade()
            grid_cost = self.communication.grid_cost
            # Each agent computes its objective power
            for k in range(len(self.agents)):
                self.agents[k].update_trade_and_price(P[:,k], grid_cost)
                trade[k], min_p_mw[k], max_p_mw[k], polycost[k], grid_cost_array[k] = self.agents[k].run(warm_start)
            memory.P[t] = trade
            self.communication.update_trade(trade)
            # Update the network and perform PF and OPF
            self.update_network(trade, min_p_mw, max_p_mw, polycost)
            loading_percent, grid = self.runpp(memory, t, grid_cost_array)
            # Compute new network cost
            grid_cost = self.SO.network_fees(loading_percent)
            memory.grid_cost[t] = grid_cost
            # Send network cost to the communication module
            self.communication.save_network_fees(grid_cost)
            self.communication.broadcast()
            pbar.update(1)
        pbar.close()


# -*- coding: utf-8 -*-
"""
Created on Wed Nov 11 10:47:36 2020

@author: Guenole CHEROT
"""
import numpy as np
import matplotlib.pyplot as plt

from scipy.optimize import Bounds
import osqp
from scipy import sparse
from copy import deepcopy


#############################
#### A JOUTER A CE CODE #####
#############################
"""
Attribu :
    - Fréquence de mise à jour du prix

fonction :
    - agent.warm_copy(max_it) # créer une copie de l'agent mais dont la série temporel est constante

"""

#############################

class Agent():
    
    ########################################################################################################
    ############################################# INIT and  SETERS #########################################
    ########################################################################################################
    
    """Agent which are part of the network. \n
    Input :
        - name (str) : Name of the agent
        - ID (int) : Unique id of the agent
        - element (int) : ID used in panda_power
        - et (str) : Type of the agent ('load', 'gen', 'ext_grid')
        - bus (int) : connected to bus n
        - n_agent (int) : Total number of agent on the grid
        - grid_cost_type (str,'uniq') : type of grid cost imposed by the SO. Can be : 'uniq', 'zone'
        - zone (int) : zone of the agent
        - flexibility (float) : How much this agent is flexible (for now the flexibility is fixed throughout the simulation)
        - rho (float) : Penalty factor of the ADMM (can be set manually for better convergence or privacy.)
        - time_series (DataFrame) : Power consumed by the agent during the simulation.
        - partner (array) : list of other agent ID this agent is allowed to trade with.
        """
    def __init__(self, name, ID, element, et, bus, n_agent, grid_cost_type, zone, flexibility, rho, time_series,
                 partners="all"):
        self.name = name
        self.ID = ID
        self.element = element
        self.et = et
        self.bus = bus
        self.n_agent = n_agent
        self.grid_cost_type = grid_cost_type
        self.zone = zone
        self.flex = flexibility
        self.rho = rho
        self.time_series = time_series
        self.partners = partners
        if self.partners == "all":
            self.partners = np.linspace(0,self.n_agent-1, self.n_agent, dtype=int)

        # time
        self.t = 0
        # power proposed by the agent
        self.P1 = np.zeros(self.n_agent)
        # counter proprosition of power by the other agents
        self.P2 = np.zeros(self.n_agent)
        # Dual variable of the problem
        self.dual = np.zeros(self.n_agent)
        # Set all bounds and poly_cost
        self.update_current_values()
        # Init grid cost at 0
        self.grid_cost = 0
        
    
    ########################################################################################################
    ########################################### Update Functions ###########################################
    ########################################################################################################
    
    def update_current_values(self):
        # Update the objective power
        self.p = 1e-3*self.time_series[self.time_series["time"]==self.t]["grid"].values[0] # multiply by 1e-3 because the power is in kW
        # Update the bound for this agent
        self.update_bounds()
        # Update the bound taking parteners into account
        self.update_bounds_partners()
        # Update the polycost based on the flexibility and objective power
        self.update_poly_cost()
    
    def update_bounds(self):
        p = self.p
        
        self.min_q_mvar = 0
        self.max_q_mvar = 0
       
        if self.et == "load":
            if p < 0 :
                self.min_p_mw = 2*p # p*(1+1/flex)
                self.max_p_mw = 0 # min(0, p*(1-1/flex))
            if p >= 0 :
                self.min_p_mw = 0 # p*(1-1/flex) # max(0, p*(1-1/flex))
                self.max_p_mw = 2*p # p*(1+1/flex)
        elif self.et == "ext_grid":
            self.min_p_mw = -5
            self.max_p_mw = 5
        else :
            raise ValueError("{0} is not in ['ext_grid','gen','load']")
    
    def update_bounds_partners(self):
        """Set linar bounds on p_nm for the agent n"""
        Min = np.zeros(self.n_agent) # Minimum power that can be exchange with other agents
        Max = np.zeros(self.n_agent) # Maximum power that can be exchange with other agents
        # Bounds on p_nm, self.partners represents all partners the agent can exange with (if no partners, Min = [0, ...,0] and  Max = [0, ...,0], no exchanges possible)
        Min[self.partners] = -np.inf
        Max[self.partners] = np.inf
        self.bounds = Bounds(Min,Max)
        
    def update_poly_cost(self):
        """Compute the polycost associated with the desired power of the agent.
        Definition :
            - polycost (array) : array constaining coeff of the cost function (f = a + b*x + c*x**2)"""
        p = self.p
        flex = self.flex
        self.polycost = [0,-2*p*flex,flex]
        
    ########################################################################################################
    ############################################# Build gamma ##############################################
    ########################################################################################################

    def build_gamma(self):
        """Build the matrix gamma representing the distance between each agent.
        Gamma is size (n_agent,n_agent).
        Input :
            - grid_cost_type (str,'uniq') : type of grid cost imposed by the SO. Can be : 'uniq', 'zone'
            - mode (str, 'Thevenin') : Method to compute the electrical distance in the network. Two electrical distances are possible : 'Thevenin' or 'PTDF'
        """
        # sign before gamma (so that the product gamma*p_nm is always positive)
        if self.p < 0:
            sign = -1
        else :
            sign = 1
        
        if self.grid_cost_type == 'uniq' :
            gamma = sign*np.ones(self.n_agent)
        if self.grid_cost_type == 'zone':
            gamma = np.zeros(self.n_agent)
            for m in range(self.n_agent):
                gamma[m] = sign*self.zone(self.ID, m)
        
        # Take the grid price into account
        self.gamma = self.grid_cost*gamma
    
    def zone(self,n,m):
        """return if the agent n and m are in the same zone. \n
        Input :
            - n (int) : agent n
            - m (int) : agent m
            
        Output :
            - 1 if the agents are in the same zone
            - 0 if not
            """
        raise ValueError("To be implemented")
        
    ########################################################################################################
    ################################### Communication with panda network ###################################
    ########################################################################################################
    
    def send_polycost_to_grid(self):
        """Set a new value to the polycost. (change the preference of the agent)
        The polycost are in the format of pandapower (we inverse in case of load for the ADMM)
        Output :
            - polycost (array) : array [a, b, c] constaining coeficient of the cost function (f = a + b*x + 0.5*c*x**2), with the good sign for panda_power"""
        polycost = deepcopy(self.polycost)
        if polycost[1] < 0 : # Power consumed by the agent
            polycost[1] += self.grid_cost
            grid_cost = self.grid_cost
        else :
            polycost[1] -= self.grid_cost
            grid_cost = -self.grid_cost
        
        if self.et == "load":
            polycost[1] *= -1
            polycost[2] *= -1
        return polycost, grid_cost
    
    def send_bounds_to_grid(self):
        """Set a new value to the polycost. (change the preference of the agent)
        The polycost are in the format of pandapower (we inverse in case of load for the ADMM)
        Output :
            - polycost (array) : array [a, b, c] constaining coeficient of the cost function (f = a + b*x + 0.5*c*x**2), with the good sign for panda_power"""
        if self.et == "load":
            min_p_mw, max_p_mw = -self.max_p_mw, -self.min_p_mw
        else :
            min_p_mw, max_p_mw = self.min_p_mw, self.max_p_mw
        return min_p_mw, max_p_mw

    ########################################################################################################
    ############################################ Core functions ############################################
    ########################################################################################################
    
    def run(self, warm_start = False):
        """Performs one step of the ADMM process.
        Input :
            - warm_start (bool) : : Performs a warm start if True (stop the time for this agent so it performs with always the same objective power)
        Output :
            - power (float) : total power consumed be the agent
            - min_p_mw, max_p_mw, polycost : new preference of the agent (only for OPF)"""
        self.update_dual()
        self.update_current_values()
        self.minimization()
        
        # time
        if warm_start == False :
            self.t += 1
        
        # Send to network
        polycost, grid_cost = self.send_polycost_to_grid()
        min_p_mw, max_p_mw = self.send_bounds_to_grid()
        
        return self.P1, min_p_mw, max_p_mw, polycost, grid_cost
    
    def update_trade_and_price(self, trade, grid_cost):
        """Update the trades proposed by other agents and network fees. \n
        Input :
            - trade (array) : array of shape (n_agent) representing the trade wanted by all other agents.
            - grid_cost (float) : cost imposed by the DSO."""
        self.P2 = trade
        self.grid_cost = grid_cost
        if self.et == "ext_grid" :
            self.grid_cost = 0
        self.build_gamma()
    
    def update_dual(self):
        """Update the Dual variable of the optimisation by looking at trades proposed by other agents. \n
        """
        self.dual -= (self.rho/2)*(self.P1+self.P2)
        
    def minimization(self):
        """Return the matrix that minimises the cost for the agent n. \n
        Output :
            - P (array) : Matrix size (n_agent,n_agent) representing the power trade between the agent n and the other agents"""  
        N = self.n_agent
        # Function to minimize
        b, c = self.polycost[1:3]
        Z = 0.5*(self.P1-self.P2)
        # Matrix for OSQP
        P = self.rho*sparse.eye(N) + 2*c*sparse.csr_matrix(np.ones((N,N)))
        q = self.gamma - self.dual - self.rho*Z + b*np.ones(N)
        A = sparse.vstack([sparse.csc_matrix(np.ones(N)),sparse.eye(N)], format='csc')
        l = np.hstack([self.min_p_mw,self.bounds.lb])
        u = np.hstack([self.max_p_mw,self.bounds.ub])
        m = osqp.OSQP()
        m.setup(P=P, q=q, A=A, l=l, u=u, verbose=False)
        results = m.solve()
        self.P1 = results.x
        return results.x

    def n_simu(self):
        """Return the number of time step of the simulation."""
        return(self.time_series.time.size)
    
    ########################################################################################################
    ################################################## PLOTS ###############################################
    ########################################################################################################    

    def __repr__(self):
        """Give usefull information about agent n."""
        # General info
        ret = "Agent n° " + str(self.ID) +"\n"
        ret += "zone : " + str(self.zone)+"\n"
        ret += "type : " + str(self.et)+"\n"
        ret += "Maximum power : " + str(self.max_p_mw)+" MW \n"
        ret += "Minimum power : " + str(self.min_p_mw)+" MW \n"
        ret += "Minimum of the cost function : " + str(-self.polycost[1]/(2*self.polycost[2])) + " MW \n"
        ret += "Flexibility : " + str(self.flex) + "\n"
        ret += "Rho : " + str(self.rho)
        return ret
        
    def plot_cost(self):
        """Plot the cost function of the agent."""
        a, b, c = self.polycost[0:3]
        if self.et == "ext_grid":
             P = np.linspace(0,1000, 100)
        else :
            min_p = -b/(2*c)
            P = np.linspace(min(2*min_p,0),max(2*min_p,0), 100)
        plt.plot(P,a + b*P + c*np.square(P))
        plt.grid()
        plt.title("Cost function of the agent {0}".format(self.ID))
        plt.xlabel("Power (MW)")
        plt.ylabel("Cost (€)")
        plt.show()
        
    def plot_gamma(self):
        """Plot the aditionnal cost impossed by the SO to the agent."""
        self.build_gamma()
        M = np.linspace(0,self.n_agent-1,self.n_agent)
        plt.bar(M, self.gamma)
        plt.title("Peer to peer cost imposed by SO to agent {0}".format(self.ID))
        plt.xlabel("agent m")
        plt.ylabel("Gamma (€/MW)")
        plt.grid()
        plt.show()
        
    def plot_bounds(self):
        """Plot the bounds of the power trades"""
        M = np.linspace(0,self.n_agent-1,self.n_agent)
        plt.bar(M, height=self.bounds.ub-self.bounds.lb, bottom=self.bounds.lb)
        plt.title("possible trade between agent {0} and agent m".format(self.ID))
        plt.xlabel("agent m")
        plt.ylabel("power trade possible (MW)")
        plt.grid()
        plt.show()
    
    def plot_power_trade(self):
        """Plot the power exanged to all the other agents."""
        # Plot the cost function 
        M = np.linspace(0,self.n_agent-1,self.n_agent)
        plt.bar(M, self.P1)
        plt.grid()
        plt.title("Power exchanged from agent {0} to agent m".format(self.ID))
        plt.xlabel("Power (MW)")
        plt.ylabel("agent m")
        plt.grid()
        plt.show()
        


# -*- coding: utf-8 -*-
"""
Created on Wed Dec  2 15:49:33 2020

@author: Guenole CHEROT
"""

class SO():
    """System Operator of the network.
    He is able to change the price of each transaction in order to change agents consumption and line loading percent. \n
    Input :
        - n_agent (int) : number of agent in the network
        - grid_cost (float) : unit cost used be the SO the regulate the market (only used when "strategy == None")
        - strategy (str, "None") : tarification strategy. Can be : "None", "PI"
        - Kp (float, 1) : proportional gain for the correction of the price (see self.PI_tarification)
        - Ki (float, 1) : integral gain for the correction of the price (see self.PI_tarification)
        - Kd (float, 1) : proportionnal gain for the correction of the price (see self.PI_tarification)
        - Ts (float, 1) : Sample time for the correction of the price (see self.PI_tarification)
        - line_limits (float, 40) : line maximum loading in percent.
        - grid_cost_type (str,'uniq') : type of grid cost imposed by the SO. Can be : 'uniq', 'zone'
    """
    
    ########################################################################################################
    ########################################### INIT and  SETERS ###########################################
    ########################################################################################################
    
    def __init__(self, n_agent, grid_cost, strategy="None", Kp=1, Ki=1, Kd=0, Ts=1, line_limits=40, grid_cost_type="uniq"):
        self.n_agent = n_agent
        if strategy=="None":
            self.grid_cost = grid_cost
        else :
            self.grid_cost = 0
        self.strategy = strategy
        self.Kp = Kp
        self.Ki = Ki
        self.Kd = Kd
        self.Ts = Ts
        self.line_limits = line_limits
        self.grid_cost_type = grid_cost_type
        # time
        self.t = 0
        self.margin_sum = 0
        
        # Used in the PID controler (memory of the margin)
        self.margin_1, self.margin_2 = 0, 0
        self.cost_1 = None
        self.cost = 0
        
        
    ########################################################################################################
    ############################################ SET NETWORK COST ##########################################
    ########################################################################################################
    
    def network_fees(self, loading_percent):
        """
        Input :
            - loading_percent (array) : array of size (n_line) representing each line overload"""
        # Update price
        self.grid_cost = self.proposed_cost(loading_percent)
        self.t+=1
        return self.grid_cost
        
    def proposed_cost(self, line_loading_percent):
        """Cost proposed by the SO inder to change the total amount of energy exchanged in the system. \n
        Input :
            - line_loading_percent (array) : Vector of size (n_line) representing the loading percent of each line.
        Output :
            - cost (float) : new cost"""
        # The first idea is to rise the price if the line loading are above 100% (We could also implement a PID or something like that)
        if self.strategy == "None" :
            cost = self.grid_cost
        elif self.strategy == "PI":
            cost = self.PI_tarification(line_loading_percent)
        else :
            raise ValueError("The tarification {0} does not exist.".format(self.strategy))
        return cost
    
    def PI_tarification(self, line_loading_percent):
        """Compute the the price with a PID controler. The control variable is the maximum loading percent of all lines in the network.
        Input :
            - line_loading_percent : Loading of each line in percent."""
        # Constants for the PID
        a = self.Kp + self.Ki*self.Ts/2 + self.Kd/self.Ts
        b = -self.Kp + self.Ki*self.Ts/2 - 2*self.Kd/self.Ts
        c = self.Kd/self.Ts
        
        self.margin = max(line_loading_percent) - self.line_limits
        if self.cost <= 0:
            self.margin = max(0, self.margin)
        
        if self.cost_1 == None : #This mean this is the first iteration
            self.cost = self.Kp*self.margin + self.Ki*self.Ts*self.margin + (self.Kd/self.Ts)*(self.margin - self.margin_1)
            self.cost_1 = self.cost
            return max(0,self.cost)
        
        if self.t%1 == 0:
            self.cost = self.cost_1 + a*self.margin + b*self.margin_1 + c*self.margin_2
            self.margin_1, self.margin_2 = self.margin, self.margin_1
            self.cost_1 = self.cost
            return max(0,self.cost)
        else :
            return max(0,self.cost)
        
    ########################################################################################################
    ################################################ INFO SO ###############################################
    ########################################################################################################
    
    def __repr__(self):
        """Give usefull information about agent n."""
        # General info
        ret = "Kp : " + str(self.Kp) +"\n"
        ret += "Ki : " + str(self.Ki)+"\n"
        ret += "Kd : " + str(self.Kd)+"\n"
        ret += "Ts : " + str(self.Ts)+"\n"
        ret += "Line limits : " + str(self.line_limits)
        return ret


# __version__ = "1"

# import os
# pp_dir = os.path.dirname(os.path.realpath(__file__))

# from rt_congestion_control.admm import *
# from rt_congestion_control.agent import *
# from rt_congestion_control.communication import *
# from rt_congestion_control.memory import *
# from rt_congestion_control.so import *
__version__ = "v0.1"

from . import admm
from . import agent
from . import communication
from . import memory
from . import so

__all__ = [
    'admm',
    'agent',
    'communication',
    'memory',
    'so'
]


import setuptools
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    README = readme_file.read()

with open('HISTORY.md') as history_file:
    HISTORY = history_file.read()

setup_args = dict(
    name='RT-Congestion-Control',
    version='1.1',
    description='Implement exogenous peer to peer algorithm for Real time Congestion Control in distribution networks',
    long_description_content_type="text/markdown",
    long_description=README + '\n\n' + HISTORY,
    license='MIT',
    packages=find_packages(),
    author='Guénolé CHEROT',
    author_email='guenole.cherot@ens-rennes.fr',
    keywords=['Distributed Optimisation', 'Peer-to-peer', 'Distribution System Operator',
     'Cost Allocation','Real-Time', 'Flexibility', 'Congestion Management'],
    url='https://gitlab.com/satie.sete/rt-congestion-control',
    download_url='https://gitlab.com/Guenole.cherot/admm/-/archive/1.1/admm-1.1.tar.gz')


install_requires = [
    'pandapower',
    'flake8',
    'matplotlib',
    'osqp',
    'tqdm',
    'numpy',
    'scipy',
    'joblib',
    'numba'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)


# -*- coding: utf-8 -*-
"""
Created on Mon Jan 11 14:46:32 2021

@author: Guenole CHEROT
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import colors
import statistics

class Memory():
    """Memory of all data acquired during simulation :
        - Dual update
        - Residuals
        - Power exchanged though Exogenous algo, OPF, OPF without constrains
        - Network fees (gamma)
        - network state (line rating, bus voltage)
        
    Input :
        - n_agent (int) : number of agent in the network (including the ext grid)
        - n_step (int) : number of time step
        - n_line (int) : number of line in the network
        - n_bus (int) : number of bus in the network
        - grid_cost_type (str,'uniq') : type of grid cost imposed by the SO. Can be : 'uniq', 'dist', 'zone'
        - line_limits (float) : line maximum loading in percent
        - compute_opf (int, 1) : 0 does not compute OPF, 1 compute OPF with line limits, 2 compute OPF with and without line limits
        - data (DataFrame) : times serie of the power desired by each agent
        - flex_data (DataFrame) : Flexibility of each agent
        """
    
    def __init__(self, n_agent, n_step, n_line, n_bus, grid_cost_type, line_limits, compute_opf, data, flex_data):
        self.n_agent = n_agent
        self.n_step = n_step
        self.n_line = n_line
        self.n_bus = n_bus
        
        self.grid_cost_type = grid_cost_type
        self.data = data
        self.flex_data = flex_data
        
        self.dual = np.zeros((n_step, n_agent, n_agent))
        self.P = np.zeros((n_step, n_agent, n_agent))
        self.grid_cost = np.zeros(n_step)
        self.line_limits = line_limits
        self.compute_opf = compute_opf
        self.line_loading_percent = np.zeros((n_step, n_line))
        self.P_exchanged = np.zeros((n_step, n_agent))
        # self.Q_exchanged = np.zeros((n_step, n_agent))
        
        if self.compute_opf >=1 :
            self.line_loading_percent_opf = np.zeros((n_step, n_line))
            self.P_exchanged_opf = np.zeros((n_step, n_agent))
            # self.Q_exchanged_opf = np.zeros((n_step, n_agent))
        if self.compute_opf == 2 :
            self.line_loading_percent_opf_2 = np.zeros((n_step, n_line))
            self.P_exchanged_opf_2 = np.zeros((n_step, n_agent))
            # self.Q_exchanged_opf_2 = np.zeros((n_step, n_agent))
        
        self.bus_voltage = np.zeros((n_step, n_bus))
        self.prim_res = np.zeros(n_step)
        self.dual_res = np.zeros(n_step)
        
    def opf_results(self, loading_percent_opf, P_exchanged_opf, k, t):
        """Store the OPF results.
        Input :
            - loading_percent_opf (array) : loading of each line in the network
            - P_exchanged_opf (array) : power exchanged by each agent in the network
            - k (int) : If 0, save in OPF. If 1 save in OPF without limits
            - t (int) : time of the simulation.
            """
        if k == 0 :
            self.line_loading_percent_opf[t] = loading_percent_opf
            self.P_exchanged_opf[t] = P_exchanged_opf
        if k == 1 :
            self.line_loading_percent_opf_2[t] = loading_percent_opf
            self.P_exchanged_opf_2[t] = P_exchanged_opf
        
    def compute_residual(self):
        """Compute the residuals"""
        for k in range(self.n_step):
            self.prim_res[k] = np.sum(np.sum(np.square(self.P[k] + self.P[k].T), axis=1))
            self.dual_res[k] = np.sum(np.sum(np.square(self.P[k] - self.P[k-1]), axis=1))
        self.prim_res[0] = self.prim_res[1]
        self.dual_res[0] = self.dual_res[1]
        return self.prim_res, self.dual_res

    ######################################################################################################
    ############################################### Stats ################################################
    ######################################################################################################
    
    def stats_power(self):
        """Return the median and last 5% decile of the difference between the power exchanged trough the exogenous algo and the OPF for each agent."""
        if self.compute_opf == 0:
            raise ValueError("OPF as not been computed during simulation. compute_opf = 0.")
        Delta_P = 100*np.sum(np.abs(self.P_exchanged - self.P_exchanged_opf), axis=0)/np.sum(np.abs(self.P_exchanged_opf), axis=0)
        q = statistics.quantiles(Delta_P, n = 20)
        median = q[9]
        last_5_percent = q[-1]
        return median, last_5_percent
    
    def stats_lines(self, compare_opf=False):
        """Return the median and last 5% decile of the line overloading. Difference between exogenous algo and line limits for every time step.
        Input :
            - compare_opf (bool, False) : if True the line loading will be compared with the opf. With self.line_limits otherwise."""
        l = np.max(self.line_loading_percent, axis=1)
        if compare_opf:
            if self.compute_opf == 0:
                raise ValueError("OPF as not been computed during simulation. compute_opf = 0.")
            l_opf = np.max(self.line_loading_percent_opf, axis=1)
            overload = l-l_opf
        else :
            overload = l-self.line_limits
        overload[overload<0] = 0
        q = statistics.quantiles(overload, n = 20)
        median = q[9]
        last_5_percent = q[-1]
        return median, last_5_percent
        
    ######################################################################################################
    ################################################ PLOT ################################################
    ######################################################################################################
    
    def plot_convergence(self, scale = False, ax=None, title=False):
        """Plot evolution of the primal and dual residual depending on the number of iteration. \n
        Input :
            - scale (bool, False) : Set to True to plot : (primal_residual/sum(Power^2)).
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True"""
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots(figsize=(6.4, 2.5), dpi=300)
        
        prim, dual = self.compute_residual()
        if scale :
            scale_factor = np.sum(np.sum(np.square(self.P), axis=1), axis=1)
            label = "(in %)"
            # Divide by 100 because the result is in %
            prim /= scale_factor/100
            dual /= scale_factor/100
        else :
            label = ""
        ax.semilogy(prim, label="Primal residual")
        ax.semilogy(dual, label="Dual residual")
        ax.grid()
        ax.legend()
        if title:
            ax.set_title("Evolution of residuals troughout the simulation")
        ax.set_xlabel("Time (min)")
        ax.set_ylabel("Residual " + label)
        ax.set_xlim((0,self.n_step))
        if plot :
            plt.show()
        
    def plot_agent_power_during_simu(self, list_agent=None, legend=False, ax=None, title=False):
        """Plot the power calculated by each agent by the algorithm throughout the simulation.
        Input :
            - list_agent (array) : list of agent ID to be ploted.
            - legend (bool) : If True, display the legend
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True
        """
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots()
        
        if list_agent is not None :
            for k in list_agent :
                ax.plot(np.sum(self.P[:,k,:],axis=1), label="agent {0}".format(k))
        else :
            for k in range(self.n_agent):
                ax.plot(np.sum(self.P[:,k,:],axis=1), label="agent {0}".format(k))
        if legend == True :
            ax.legend()
        ax.grid()
        if title:
            ax.set_title("Power computed during the simulation")
        ax.set_xlabel("time (min)")
        ax.set_ylabel("Power (MW)")
        ax.set_xlim((0,self.n_step))
        if plot :
            plt.show()
        
    def plot_compare_to_OPF_simu(self, ID_agent, legend=False, objective=False, ax=None, title=False):
        """Plot the power calculated by agent "ID_agent" troughout the simulation.
        Input :
            - ID_agent : agent ID to be ploted.
            - legend (bool) : If True, display the legend
            - objective (bool) : If True, display the objective power
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True
        """
        plot = False
        name = ["(a) ", "(b) ", "(c) ", "(d) "]
        if ax == None:
            plot = True
            name = ["","","",""]
            fig, ax = plt.subplots()
        
        ex, = ax.plot(1e3*self.P_exchanged[:,ID_agent], label="{0}Exogenous".format(name[0]))
        
        if self.compute_opf >= 1:
            opf, = ax.plot(1e3*self.P_exchanged_opf[:,ID_agent], color='orange', label="{0}OPF".format(name[2]))
        if self.compute_opf == 2:
            opf_2, = ax.plot(1e3*self.P_exchanged_opf_2[:,ID_agent],'--', color='xkcd:orange', label="{0}OPF without line limits".format(name[3]))
        
        if objective :
            p = self.data[self.data["dataid"]==ID_agent].sort_values(by='time').grid.values
            obj, = ax.plot(p, '--g', label="{0}Objective".format(name[1]))

        if legend == True :
            ax.legend()
        ax.grid()
        if title:
            ax.set_title("Power produced during the simulation, f={0}".format(np.round(self.flex_data.flexibility[ID_agent],0)))
        ax.set_xlabel("time (min)")
        ax.set_ylabel("Power (kW)")
        ax.set_xlim((0,self.n_step))
        
        ax.text(0.05, 0.15, "Agent {0}".format(ID_agent), transform=ax.transAxes,
                fontsize=10, verticalalignment='top')
        
        if plot :
            plt.show()
        else :
            if self.compute_opf == 2:
                return ex, obj, opf, opf_2
            else :
                raise ValueError("compute_opf = 2 mandatory")
            
    def plot_line_loading(self, t, ax=None, title=False):
        """Plot the line loading resulting from the load flow.\n
        Input :
            - t (int) : time of the simulation.
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True
        """
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots()
        M = np.linspace(0,self.n_line-1,self.n_line)
        ax.bar(M, height=self.line_loading_percent[t], bottom=0)
        ax.set_ylim((0,200))
        ax.set_xlim((-.5,self.n_agent+0.5))
        ax.plot(M, 100*np.ones(len(M)), 'g--')
        if title:
            ax.set_title("loading percent of every line")
        ax.set_xlabel("Line n")
        ax.set_ylabel("Loading (%)")
        ax.grid()
        if plot :
            plt.show()
            

    def plot_cost_evolution(self, ax=None, title=False):
        """Plot the evolution of the cost during one simulation. \n
        Input :
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True"""
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots()
        
        T = np.linspace(0,self.n_step-1,self.n_step)
        if self.grid_cost_type == "uniq":
            ax.plot(T, 60*self.grid_cost, color="black") # We multiply by 60 to have the price in €/MWh and not €/MWmin
            if title:
                ax.set_title("Evolution of the cost during the simulation")
            ax.set_xlabel("time (min)")
            ax.set_ylabel("Grid cost (€/MWh)")
            ax.set_xlim((0,self.n_step))
            ax.grid()
            if plot :
                plt.show()
        else :
            raise ValueError("Not implemented yet")

    def plot_max_loading_percent(self, ax=None, title=False):
        """Plot the maximum loading percent during a simulaiton. \n
        Input :
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True
        """
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots()
        
        l = np.max(self.line_loading_percent, axis=1)
        T = np.linspace(0,self.n_step-1,self.n_step)
        ax.plot(T, l, label="Exogenous")
        
        if self.compute_opf >= 1:
            l_opf = np.max(self.line_loading_percent_opf, axis=1)
            ax.plot(T, l_opf,color='orange', label="OPF")
            
        if self.compute_opf == 2:
            l_opf_2 = np.max(self.line_loading_percent_opf_2, axis=1)
            ax.plot(T, l_opf_2,'--',color='xkcd:orange', label="OPF without line limits")
        
        if title:
            ax.set_title("Evolution of the maximum loading during the simulation")
        ax.set_xlabel("time (min)")
        ax.set_ylabel("Maximum loading of the lines (in %)")
        ax.set_xlim((0,self.n_step))
        # ax.set_ylim(55,81)
        ax.legend()
        ax.grid()
        if plot :
            plt.show()
        
    def plot_all_agent_power(self, t, warm=False, ax=None, title=False):
        """Plot the power exchanged with the exogenous algorithm and with the OPF at time t.
        Input :
            - t (int) : time slot to visualise.
            - warm (bool) : If True plot the result of the warm start.
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True"""
        if warm :
            desired = 1e-3*self.data[self.data["time"]==0].sort_values(by='dataid')["grid"].values
        else :
            desired = 1e-3*self.data[self.data["time"]==t].sort_values(by='dataid')["grid"].values
        # Power computed by the power flow
        exchanged = np.copy(self.P_exchanged[t])
        if self.compute_opf >= 1 :
            # Power computed by the OPTIMAL power flow
            exchanged_opf = np.copy(self.P_exchanged_opf[t])
        
        x = np.arange(self.n_agent)  # the label locations
        width = 0.2  # the width of the bars
        
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots()
        rects1 = ax.bar(x - width, desired, width, label="P desired")
        rects2 = ax.bar(x, exchanged, width, label="P exogenous")
        if self.compute_opf >= 1 :
            rects3 = ax.bar(x + width, exchanged_opf, width, label ="P OPF")
        ax.set_ylabel('Power consumed (MW)')
        ax.set_xlabel("agents number")
        if title:
            ax.set_title("Power exchanged by each agent at time step {0}".format(t))
        ax.set_xticks(x)
        ax.legend()
        if plot :
            plt.show()
        
    def plot_cost_objective_function(self, t, verbose=False, ax=None, title=False):
        """Compute the grid cost paid by each agent at t. \n
        Input :
            - t (int) : time of the simulation
            - verbose (bool, True) : return total cost if True
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True"""
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots()
        
        p = 1e-3*self.data[self.data["time"]==t].sort_values(by='dataid')["grid"].values
        res_p = np.copy(self.P_exchanged[t])
        f = self.flex_data.flexibility
        a = f*p*p
        b = -2*f*p
        
        cost = a + b*res_p + f*res_p*res_p
        cost_tot = np.sum(cost)
        
        # PLOT
        M = np.linspace(0,self.n_agent-1,self.n_agent)
        ax.bar(M, height=cost, bottom=0)
        ax.set_xlim((-.5,self.n_agent+0.5))
        if title:
            ax.set_title("Cost suffered by each agent a time {0}".format(t))
        ax.set_xlabel("Agent n")
        ax.set_ylabel("cost objective function")
        ax.grid()
        if plot :
            plt.show()
        if verbose :
            return (cost_tot)
    
    def plot_flexibility(self, ax=None, title=False):
        """Compute the total cost t the end of the simulation. \n
        Input :
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True"""
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots()
        
        M = np.linspace(0,self.n_agent-1,self.n_agent)
        ax.bar(M, height=self.flex_data.flexibility, bottom=0)
        ax.set_xlim((-.5,self.n_agent+0.5))
        if title:
            ax.set_title("Flexibility of each agent")
        ax.set_xlabel("Agent n")
        ax.set_ylabel("Flexibility")
        ax.grid()
        if plot :
            plt.show()
    
    def plot_fig1(self, ID_not_flex, ID_flex, title=False):
        """Plot the first fgure in the article ISGT.
        Input :
            - ID_not_flex (int) : ID of a non flexible agent.
            - ID_flex (int) : ID of a flexible agent.
            - title (bool, False) : print figure's title if True"""
        if self.compute_opf != 2:
            print("Can't plot fig, compute_opf needs to be egals to 2.")
            return(self.compute_opf)
        fig = plt.figure(figsize=(8,0.8*8), dpi=600)
        gs = fig.add_gridspec(4, hspace=0)
        axs = gs.subplots(sharex=True)
        ex, obj, opf, opf_2 = self.plot_compare_to_OPF_simu(ID_agent=ID_not_flex, legend=False, objective=True, ax=axs[0])
        self.plot_compare_to_OPF_simu(ID_agent=ID_flex, legend=False, objective=True, ax=axs[1])
        self.plot_max_loading_percent(ax=axs[2])
        self.plot_cost_evolution(ax=axs[3])
        # # Common y axis for the 2 first plot
        # axs[0].sharey(axs[1])
        # legend for the 3 firts pots
        first_legend = axs[0].legend(handles=[ex, obj], loc=(0.36, -0.2), framealpha=1)
        axs[0].set_zorder(1)
        axs[0].set_xlabel("")
        # Add the legend manually to the current Axes.
        axs[0].add_artist(first_legend)
        # Create another legend for the second line.
        axs[0].legend(handles=[opf, opf_2], loc=(0.62, -0.2), framealpha=1)
        axs[2].get_legend().remove()
        
        # Delete y label and title
        for k in range(4):
            axs[k].set_title("")
            axs[k].set_ylabel("")
            
        if title:
            fig.suptitle('Evolution of the power, cost and line loading during one simulation', fontsize=14, x=0.5,y = 0.93)
        # ylabel
        x = 0.04
        fig.text(x+0.01, 0.79, 'Power (kW)', va='center', rotation='vertical')
        fig.text(x+0.01, 0.6, 'Power (kW)', va='center', rotation='vertical')
        fig.text(x, 0.42, 'Max line \n loading (%)', va='center', rotation='vertical')
        fig.text(x, 0.2, 'Cost \n (€/MWh)', va='center', rotation='vertical')

        plt.show()
        
    def plot_cost_vs_flex(self, ID_not_flex=None, ID_flex=None, ax=None, title=False):
        """Plot the grid cost paid by each agent as a function of their fllexibility. \n
        Input :
            - ID_not_flex (int, None) : ID of an unflexible agent. Default is None, not plotted.
            - ID_flex (int, None) : ID of a flexible agent. Default is None, not plotted.
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True"""
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots(figsize=(6.4, 2.5), dpi=300)
        
        exchanged = np.abs(self.P_exchanged)
        cost = np.matmul(self.grid_cost,exchanged)
            
        cost = np.divide(cost,np.sum(exchanged, axis=0)/60)
        ax.plot(self.flex_data.flexibility[1:], cost[1:], "o")
        if title:
            ax.set_title("Grid cost paid function of flexibility")
        ax.set_ylabel("Grid cost paid (€/MWh)")
        ax.set_xlabel("Flexibility")
        ax.grid()
        
        # Spot specific agent
        if ID_not_flex != None :
            x, y = (self.flex_data.flexibility[ID_not_flex], cost[ID_not_flex])
            ax.annotate('Agent {0}'.format(ID_not_flex), xy=(x, y),  xycoords='data',
                xytext=(x-10, y-6), textcoords='data',
                arrowprops=dict(facecolor='green', shrink=0.05),
                horizontalalignment='right', verticalalignment='top')
            
        if ID_flex != None :
            x, y = (self.flex_data.flexibility[ID_flex], cost[ID_flex])
            ax.annotate('Agent {0}'.format(ID_flex), xy=(x, y),  xycoords='data',
                xytext=(x+50, y-4), textcoords='data',
                arrowprops=dict(facecolor='green', shrink=0.05),
                horizontalalignment='right', verticalalignment='top')
            
        if plot:
            plt.show()
            
    def plot_overload_histogram(self, ax=None, title=False):
        """Plot histogram of all line overload throughtout the simulation. \n
        Input :
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot
            - title (bool, False) : print figure's title if True"""
        plot = False
        if ax == None:
            plot = True
            fig, ax = plt.subplots(figsize=(6.4, 2.1), dpi=300)
        
        m1 = self.line_loading_percent.max()
        m2 = self.line_loading_percent_opf_2.max()
        m = max(m1,m2)
        bins = np.linspace(0,int(m)+1,int(m)+2)
        n,_,_ = ax.hist(self.line_loading_percent.flatten(), bins=bins, label="Exogenous")
        n2,_,_ = ax.hist(self.line_loading_percent_opf_2.flatten(), label="OPF without constrains", bins=bins, alpha=0.5, zorder=0)#, histtype=u'step', linewidth=1)
        oc = max(n)
        ax.plot([self.line_limits,self.line_limits], [0, oc*1.02], '--g')
        if title:
            ax.set_title("Histogram of the overload")
        ax.set_ylabel("Number of occurence")
        ax.set_xlabel("Line loading (%)")
        ax.set_xlim((0,int(m)+1))
        # ax.set_ylim((0,oc*1.02))
        ax.set_yscale('log', nonposy='clip')
        ax.legend(framealpha=1)
        ax.grid()
        
        ax.annotate('maximum  \n line loading', xy=(self.line_limits, oc/20),  xycoords='data',
            xytext=(self.line_limits*0.83, oc/2), textcoords='data',
            arrowprops=dict(facecolor='green', shrink=0.05),
            horizontalalignment='right', verticalalignment='top')
            
        if plot:
            plt.show()
        print(round(100*np.sum(n[100:])/np.sum(n),1),"% above 100%.")
        print("Maximum loading = ", np.size(n), "%")
            
    def PI_sensi_analysis(self, Stats_power, Stats_lines, Kp, Ki, custom_label = False, title=False):
        """Plot the sensivity analysis for the PI controller.
        Input :
            - Stats_power (array) : Array of shape (n_Kp, n_Ki, 2) representing the median and last decile for the power exchange during the simulation.
            - Stats_lines (array) : Array of shape (n_Kp, n_Ki, 2) representing the median and last decile for the line overloading during the simulation.
            - Kp (array) : Kp (proportional gain) array used to compute Stat_power and Stat_lines.
            - Ki (array) : Ki (integral gain) array used to compute Stat_power and Stat_lines.
            - custom_label (bool, False) : Set the custom label if true. Automatic otherwise.
            - ax (plt ax, None) : matplotlib pyplot ax on wich to plot.
            - title (bool, False) : print figure's title if True.
            """
        fig, axs = plt.subplots(2, 2, figsize=(7, 7), dpi=300, sharex=True, sharey=True)
        if title:
            fig.suptitle('PI sensivity analysis')
        
        Stats_lines[Stats_lines==0] = 0.001
        
        images = []
        for i in range(2):
            for j in range(2):
                if i == 0 :
                    data = Stats_power[:,:,j]
                else :
                    data = Stats_lines[:,:,j]
                images.append(axs[i, j].imshow(data, origin="lower"))
                axs[i, j].label_outer()
        
        # Set title
        axs[0, 0].set_title('(a) Median power not delived')
        axs[0, 1].set_title('(b) 95% quantile \n power not delived')
        axs[1, 0].set_title('(c) Median overflow')
        axs[1, 1].set_title('(d) 95% quantile overflow')
        
        # Set axis label
        if custom_label :
            axs[0, 0].set_xticks([0,1,7,13])
            axs[0, 0].set_yticks([0,1,5,9])
            axs[0, 0].set_xticklabels(["0", "1e-3", "0.01", "0,1"])
            axs[0, 0].set_yticklabels(["0", "0.001", "0.01", "0,1"])
        else :
            axs[0, 0].set_xticks(np.arange(len(Ki)))
            axs[0, 0].set_yticks(np.arange(len(Kp)))
            axs[0, 0].set_xticklabels(Ki)
            axs[0, 0].set_yticklabels(Kp)
    
        axs[0,0].set(xlabel='', ylabel='Kp')
        axs[0,1].set(xlabel='', ylabel='')
        axs[1,0].set(xlabel='Ki', ylabel='Kp')
        axs[1,1].set(xlabel='Ki', ylabel='')
            
        # Find the min and max of all colors for use in setting the color scale.
        vmin = 1 #min(np.min(Stats_power),np.min(Stats_lines))
        vmax = max(np.max(Stats_power),np.max(Stats_lines))
        norm = colors.LogNorm(vmin=vmin, vmax=vmax)
        for im in images:
            im.set_norm(norm)
        
        # Change position of the plots
        plt.subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.05, hspace=0)
        x,y,w,h = axs[0,0].get_position().bounds
        axs[0,0].set_position([x,y-0.05,w,h])
        x,y,w,h = axs[0,1].get_position().bounds
        axs[0,1].set_position([x,y-0.05,w,h])
        # fig.tight_layout()
        
        cax = plt.axes([0.1, +0.05, 0.8, 0.04])
        cbar = fig.colorbar(images[0],cax=cax, orientation='horizontal', ticks=[1,5, 10, 20])
        cbar.ax.set_xticklabels(['<1%','5%', '10%', '20%'])
        plt.show()
        